/*-
 * Copyright (C)2003 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C)2003 @BABOLO http://www.babolo.ru/\n\
@(#)All rights reserved.\n";
static const char rcsid[] = "$Id: open.c,v 1.1 2003/11/26 04:07:05 babolo Exp $";
#endif /* not lint */

#include <sys/types.h>
#include <sys/mman.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <stdio.h>
#include "mife_int.h"
#include "mife.h"

struct mife_descriptor *
mife_open(const char *path, u_int32_t flag) {
    struct mife_descriptor *mife;

    if  (!(mife = malloc(sizeof(struct mife_descriptor)))) {
        errno = ENOMEM;
        return(NULL);
    }
    bzero(mife, sizeof(struct mife_descriptor));
    mife->flag = flag & ~MIFE_CLOS;
    mife->mindelta = MINDELTA;
    if  ((mife->fd = open(path, O_RDONLY)) < 0) return(NULL);
    if  ((mife->flag & MIFE_RTRY) == 0) mife->flag |= NRETRY00 & MIFE_RTRY;
    mife->flag &= ~MIFE_INMM;
    return(mife);
}

struct mife_descriptor *
mife_opef(int fd, u_int32_t flag) {
    struct mife_descriptor *mife;

    if  (!(mife = malloc(sizeof(struct mife_descriptor)))) {
        errno = ENOMEM;
        return(NULL);
    }
    bzero(mife, sizeof(struct mife_descriptor));
    mife->flag = flag | MIFE_CLOS;
    mife->mindelta = MINDELTA;
    mife->fd = fd;
    if  ((mife->flag & MIFE_RTRY) == 0) mife->flag |= NRETRY00 & MIFE_RTRY;
    mife->flag &= ~MIFE_INMM;
    return(mife);
}

struct mife_descriptor *
mife_opes(u_int32_t flag, const char *src, size_t size) {
    struct mife_descriptor *mife;
    size_t len;
    int i;

    if  (!(mife = malloc(sizeof(struct mife_descriptor)))) {
        errno = ENOMEM;
        return(NULL);
    }
    bzero(mife, sizeof(struct mife_descriptor));
    mife->flag = flag | MIFE_SDUP | MIFE_INMM;
    mife->mindelta = MINDELTA;
    mife->fd = -1;
    mife->offset = 0;
    len = size ? size : strlen(src);
    i = (flag & MIFE_FULL) ? 1 : 0;
    mife->len = len;
    if  (!(mife->buffer = malloc(mife->len + i))) {
        errno = ENOMEM;
        return(NULL);
    }
    bcopy(src, mife->buffer, mife->len);
    if  (i) ((char *)(mife->buffer))[mife->len] = 0;
    mife->buflen = mife->len + i;
    return(mife);
}

struct mife_descriptor *
mife_opem(u_int32_t flag, char *src, size_t size) {
    struct mife_descriptor *mife;
    size_t len;
    int i;

    if  (!(mife = malloc(sizeof(struct mife_descriptor)))) {
        errno = ENOMEM;
        return(NULL);
    }
    bzero(mife, sizeof(struct mife_descriptor));
    mife->flag = flag & ~MIFE_SDUP | MIFE_INMM;
    mife->mindelta = MINDELTA;
    mife->fd = -1;
    mife->offset = 0;
    len = size ? size : strlen(src);
    i = (mife->flag & MIFE_FULL) ? 1 : 0;
    mife->buffer = src;
    mife->len = len;
    if  (i) ((char *)(mife->buffer))[mife->len] = 0;
    mife->buflen = mife->len + i;
    return(mife);
}
