/*-
 * Copyright (C)2003 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C)2003 @BABOLO http://www.babolo.ru/\n\
@(#)All rights reserved.\n";
static const char rcsid[] = "$Id: read.c,v 1.1 2003/11/26 04:07:05 babolo Exp $";
#endif /* not lint */

#include <machine/limits.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/uio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <stdio.h>
#include "mife.h"
#include "mife_int.h"

ssize_t
mife_read(struct mife_descriptor *mife, size_t len, off_t offset) {
    struct stat sb;
    struct rlimit rlp;
    void *result;
    ssize_t tmplen;
    off_t delta;
    int retrylim;
    u_int32_t cond;
#define COND_NEED_REM  0x0001
#define COND_NEED_MOV  0x0002
#define COND_CLEAR_BUF 0x0004
#define COND_BIG_GAP   0x0008

    if  (mife->flag & MIFE_VERD) fprintf( stderr, "+mife_read %s len=%08X offset=%08X\n"
                                        , mife ? "OK" : "NULL"
                                        , len
                                        , (size_t)offset
                                        );
    if  (len >= SSIZE_MAX) {
        errno = EDOM;
        if  (mife->flag & MIFE_VERD) fprintf(stderr, "-mife_read #01: %s", strerror(errno));
        return(-1);
    }
    if  (mife->flag & MIFE_INMM) {
        if  (offset > mife->len) {
            errno = ESPIPE;
            if  (mife->flag & MIFE_VERD) fprintf(stderr, "-mife_read #02: %s", strerror(errno));
            return(-1);
        }
        tmplen = mife->len - offset;
        if  (len > 0 && len < tmplen) tmplen = len;
        if  (mife->flag & MIFE_VERD) fprintf(stderr, "-mife_read #03 null %08X\n", tmplen);
        return(tmplen);
    }
    if  (fstat(mife->fd, &sb)) {
        if  (mife->flag & MIFE_VERD) fprintf(stderr, "-mife_read #04: %s", strerror(errno));
        return(-1);
    }
    if  (mife->flag & MIFE_VERH) fprintf( stderr, "st_mode=%011o", sb.st_mode);
    switch (sb.st_mode & S_IFMT) {
    case S_IFIFO:
    case S_IFCHR:
    case S_IFBLK:
    case S_IFSOCK:
        if  (mife->flag & MIFE_VERH) fprintf( stderr, " -> pipe\n");
        if  (len == 0) {
            if  (getrlimit(RLIMIT_DATA, &rlp)) {
                if  (mife->flag & MIFE_VERD)
                    fprintf(stderr, "-mife_read #10: pipe: %s", strerror(errno));
                return(-1);
            }
            len = ((rlp.rlim_cur == RLIM_INFINITY) ? DEFMEM : rlp.rlim_cur >> 8) + mife->mindelta;
            if  (len > SSIZE_MAX) len = SSIZE_MAX;
        }
        if  (len > SSIZE_MAX) {
            errno = EINVAL;
            if  (mife->flag & MIFE_VERD) fprintf(stderr, "-mife_read #11: pipe: %s", strerror(errno));
            return(-1);
        }
        delta = offset - mife->offset;
        if  (delta < (off_t)0) {
            errno = ESPIPE;
            if  (mife->flag & MIFE_VERD) fprintf(stderr, "-mife_read #12: pipe: %s", strerror(errno));
            return(-1);
        }
        tmplen = len + ((mife->flag & MIFE_FULL) ? 1 : 0);
        tmplen = (((len > mife->mindelta) ? tmplen : mife->mindelta) + MAXPACKT) & ~(MAXPACKT - 1);
        if  (!mife->buffer) {
            cond |= COND_CLEAR_BUF;
            if  (mife->flag & MIFE_VERH) fprintf(stderr, "malloc(%d)\n", (int)tmplen);
            mife->len = 0;
            mife->buflen = tmplen;
            if  (!(mife->buffer = malloc(tmplen))) {
                errno = ENOMEM;
                if  (mife->flag & MIFE_VERD)
                    fprintf(stderr, "-mife_read #13: pipe: %s", strerror(errno));
                return(-1);
        }   }
        if  (delta >= mife->buflen) cond |= COND_BIG_GAP;
        if  (tmplen > mife->buflen) cond |= COND_NEED_REM;
        if  (cond & COND_BIG_GAP) cond |= COND_CLEAR_BUF;
        if  (cond & COND_NEED_REM && cond & COND_CLEAR_BUF) {
            cond &= ~COND_NEED_REM;
            if  (mife->flag & MIFE_VERH) fprintf(stderr, "remalloc(%d)\n", (int)tmplen);
            free(mife->buffer);
            mife->offset += mife->len;
            delta = offset - mife->offset;
            mife->len = 0;
            mife->buflen = tmplen;
            if  (!(mife->buffer = malloc(tmplen))) {
                errno = ENOMEM;
                if  (mife->flag & MIFE_VERD)
                    fprintf(stderr, "-mife_read #14: pipe: %s", strerror(errno));
                return(-1);
        }   }
        if  (mife->flag & MIFE_VERH)
            fprintf( stderr
                   , "mife->flag=%08X\nmife->fd=%d\nmife->mindelta=%08X\nmife->len=%08X\n"
                     "mife->buflen=%08X\nmife->offset=%08X\n"
                     "cond=%08X\nlen=%08X\ntmplen=%08X\noffset=%08X\ndelta=%08X\n"
                   , mife->flag
                   , mife->fd
                   , mife->mindelta
                   , mife->len
                   , mife->buflen
                   , (size_t)mife->offset
                   , cond
                   , len
                   , tmplen
                   , (size_t)offset
                   , (size_t)delta
                   );

        if  (cond & COND_BIG_GAP) {
            cond &= ~COND_BIG_GAP;
            if  (mife->flag & MIFE_VERH) fprintf(stderr, "big gap\n");
            for (retrylim = mife->flag & MIFE_RTRY; delta >= mife->buflen - len; ) {
                mife->len = (size_t)0;
                tmplen = read(mife->fd, mife->buffer, mife->buflen);
                mife->offset += mife->len;
                delta = offset - mife->offset;
                if  (tmplen < 0) {
                    if  (errno != EAGAIN && errno != EDEADLK) {
                        if  (mife->flag & MIFE_VERD)
                            fprintf(stderr, "-mife_read #15: pipe: %s", strerror(errno));
                        return(-1);
                    }
                } else if (tmplen = 0) {
                    if  (retrylim-- <= 0) {
                        errno = ESPIPE;
                        if  (mife->flag & MIFE_VERD)
                            fprintf(stderr, "-mife_read #16: pipe: %s", strerror(errno));
                        return(-1);
                    }
                    usleep(1 << (mife->flag & MIFE_RTRY - retrylim));
                } else {
                    retrylim = mife->flag & MIFE_RTRY;
                    mife->len = tmplen;
        }   }   }
        if  (mife->len < delta) cond |= COND_CLEAR_BUF;
        else if (delta + len > mife->buflen) cond |= COND_NEED_MOV;
        if  (cond & COND_NEED_MOV) {
            cond &= ~COND_NEED_MOV;
            if  (mife->flag & MIFE_VERH) fprintf(stderr, "need move\n");
            bcopy(mife->buffer + delta, mife->buffer, mife->len - delta);
            mife->offset = offset;
            mife->len -= delta;
            delta = 0;
        }
        if  (cond & COND_NEED_REM) {
            cond &= ~COND_NEED_REM;
            if  (mife->flag & MIFE_VERH) fprintf(stderr, "reallocf(%d)\n", (int)tmplen);
            if  (!(mife->buffer = reallocf(result, tmplen))) {
                errno = EFBIG;
                if  (mife->flag & MIFE_VERD) fprintf(stderr, "-mife_read #17: pipe: %s", strerror(errno));
                return(-1);
        }   }
        if  (mife->flag & MIFE_VERH) fprintf( stderr, "cond=%08X\n", cond);

        for (retrylim = mife->flag & MIFE_RTRY; len > 0; ) {
            tmplen = read(mife->fd, mife->buffer + mife->len, (len > MAXPACKT) ? len : MAXPACKT);
            if  (mife->flag & MIFE_VERH)
                fprintf( stderr, "retrylim=%d tmplen=%d len=%d\n", retrylim, tmplen, len);
            if  (tmplen < 0) {
                if  (errno != EAGAIN && errno != EDEADLK) {
                    if  (mife->flag & MIFE_VERD)
                        fprintf(stderr, "-mife_read #18: pipe: %s", strerror(errno));
                    return(-1);
                }
            } else if (tmplen == 0) {
                if  (retrylim-- <= 0) break;
                usleep(1 << (mife->flag & MIFE_RTRY - retrylim));
            } else {
                retrylim = mife->flag & MIFE_RTRY;
                mife->len += tmplen;
                len -= tmplen;
        }   }
        tmplen = mife->buflen - mife->len;
        bzero(mife->buffer + mife->len, (tmplen > 8) ? 8 : tmplen);
        if  (mife->flag & MIFE_VERD)
            fprintf(stderr, "-mife_read #19 pipe %08X\n", (ssize_t)mife->len - delta);
        return((ssize_t)mife->len - delta);
    case S_IFREG:
        if  (mife->flag & MIFE_VERH) fprintf(stderr, " -> file\n");
        if  (mife->buffer && munmap(mife->buffer, mife->buflen) < 0) {
            if  (mife->flag & MIFE_VERD) fprintf(stderr, "-mife_read #20: file: %s", strerror(errno));
            return(-1);
        }
        if  (len == 0) len = sb.st_size;
        tmplen = len + ((mife->flag & MIFE_FULL) ? 1 : 0);
        mife->buffer = mmap(NULL, tmplen, PROT_READ | PROT_WRITE, MAP_PRIVATE, mife->fd, offset);
        if  (mife->buffer == MAP_FAILED) {
            if  (mife->flag & MIFE_VERD) fprintf(stderr, "-mife_read #21: file: %s", strerror(errno));
            return(-1);
        }
        mife->len = len;
        mife->buflen = tmplen;
        mife->offset = offset;
        if  (mife->flag & MIFE_FULL) ((u_char *)(mife->buffer))[mife->len] = 0;
        if  (mife->flag & MIFE_FULL && mife->flag & MIFE_VERH)
            fprintf( stderr, "===================\n%s~~~~~~~~~~~~~~~~~~~\n", (u_char *)(mife->buffer));
        if  (mife->flag & MIFE_VERD) fprintf(stderr, "-mife_read #22 file %08X\n", tmplen);
        return(tmplen);
    case S_IFDIR:
    case S_IFLNK:
    case S_IFWHT:
    default:
        errno = EINVAL;
        if  (mife->flag & MIFE_VERD) fprintf(stderr, " -> unknown\n-mife_read #23 -1\n");
        return(-1);
}   }
